// Demo de la biblioteca grfica mgsLib.
// Implementacin de un sencillo juego de frontn.
// Germn Fabregat, 2016

// Ficheros de cabecera para las funciones de la
// biblioteca est,andar de C
#include <stdio.h>
#include <stdlib.h>

// Fichero de cabecera de mgsLib
#include "mgsLib.h"

// Definicin de constantes
#define SCRW		800			// Ancho de la pantalla
#define SCRH		600			// Alto de la pantalla
#define SCRBKGC		0x101028 	// Color de fondo 
#define TXTBKGC		0x303078	// Color del texto de fondo. Similar pero ms claro.
#define SCRTITLE  	"Demo game"	// Titulo de la ventana
#define BHWIDTH		20			// Anchura de las paredes horizontales
#define BVWIDTH		30			// Anchura de la pared vertical
#define BCOLOR  	0x5050FF	// Color de la pared
#define TITLEW		625			// Ancho del texto de inicio
#define TITLEH		90			// Alto del texto de inicio
#define TITLEX0		((SCRW - TITLEW)/2)	// Pos. inicial, centrada
#define TITLEY0		((SCRH - TITLEH)/2)
#define TITLEC		MGS_CCYAN	// Color del texto
#define BATW		60			// Ancho normal del bate
#define BATH		15			// Alto del bate
#define BATX0		((SCRW - BATW)/2)	// Pos. inicial del bate
#define BATY0		(SCRH - BATH - 10)
#define BATC		MGS_CGREEN	// Color del bate
#define BALLR		10			// Radio de la pelota
#define BALLX0		(BATX0 + BATW/2)	// Pos. inicial de la pelota
#define BALLY0		(BATY0 - BALLR) 	
#define BALLC		MGS_CBLUE	// Color de la bola
#define INCBAT 		3			// Velocidad normal del bate
#define BATWGIF		76			// Anchura del bate con mejora
#define BATVGIF		5			// Velocidad del bate con mejora
#define BATGIFC		0x60FF60	// Color del bate con mejora
#define GIFTER		20			// Radio externo de la estrella
#define GIFTIR		10			// Radio interno de la estrella
#define GIFTIXM		0			// Indice tipo de mejora
#define GIFTIXP		1			// Indice nmero de puntas
#define GIFTIXR		2			// Indice ngulo rotado
#define GIFTCW		0			// Mejora tipo cambio de anchura
#define GIFTCV		1			// Mejora tipo cambio de velocidad
#define GIFTMAX		750			// Retardo base de aparicin de estrellas
#define BATGIFT		800			// Duracin de la mejora
#define ANIMOUT		200			// Duracin de la animacin de bola perdida	(estado GOOUT)		
#define MAXFIX		10			// Duracin para el cambio de velocidad del titulo (estado PREVIOUS)
#define EXTATTR		3			// Nmero de atributos sin definir de los objetos

// Estados del bucle infinito de programa
// PREVIOUS muestra el texto y espera que se pulse S
// NORMAL es el modo de juego normal
// GOOUT nos permite insertar una animacin cuando perdemos la bola
enum states {PREVIOUS, NORMAL, GOOUT};

// Estructura que define nuestros objetos.
// x, y guardan la posicin caraterstica segn el tipo de objeto
// vx, vy la velocidad en pixeles en cada eje
// w, h la anchura y altura, segn el tipo de objeto
// c da el color.
// Pera los usos del programa se tiene un vector de caractersitcas
// propias de cada objeto.
// El uso de una estructura as da generalidad y orden, a la vez
// que permite gracias al vector particularizar para cada caso
struct object {
	int x, y, vx, vy, w, h, c;
	int extra[EXTATTR];
};

// Funcin principal y nica, sin contar las de biblioteca
// como es sencillo, no usamos funciones adicionales.
// La estructura es la tpica de un programa repetitivo y dirigido 
// por eventos de E/S, tpica de los sistemas empotrados y los juegos.
// Comenzamos con la inicializacin del sistema y luego entramos en 
// un bucle infinito temporizado que permite realizar animaciones.
// Para tener distintos modos dentro de ese bucle, usamos distintos
// estados.
int main()
{
	// Variables locales.
	// i se utiliza para contadores y d para clculos intermedios.
	// scr identifica la pantalla y err recoge los errores devueltos por las funciones.
	// animOut, giftTime y widthTime se utilizan para computar tiempos.
	int i, d, scr, err, animOut, giftTime, widthTime = 0;
	// Objetos del juego
	struct object title, ball, bat, gift, fragments[8];
	// Contadores del juego
	int out = 0, rebound = 0, nStars = 0, cStars = 0;
	// Estados de ejecucin
	enum states st = PREVIOUS;
	// Cadenas para mostrar dos contadores en la pantalla grfica.
	char sRebounds[10], sOut[10];
	
	// Creamos una pantalla de SCRWxSCRH con color de fondo. Tratamos el posible error
	if ((scr = mgsCreateScreen(SCRW, SCRH, SCRTITLE, SCRBKGC, NULL)) < 0) {
		mgsShowError(scr, "Error creating screen. Aborting program", MGS_EFWAIT);
		exit(EXIT_FAILURE);
	}
	
	// Dibujamos los bordes
	mgsRectangle(scr, 0, 0, SCRW, BVWIDTH, BCOLOR);
	mgsRectangle(scr, 0, 0, BHWIDTH, SCRH, BCOLOR);
	mgsRectangle(scr, SCRW - BHWIDTH, 0, SCRW, SCRH, BCOLOR);
	
	// Y hacemos que este dibujo sea el fondo
	if ((err = mgsSetBackground(scr)) < 0) {
		mgsShowError(err, "Error setting background. Aborting program", MGS_EFWAIT);
		exit(EXIT_FAILURE);
	}
	
	// Ponemos la temporizacion a 100 FPS
	mgsSetFrameTime(10);
	
	// Ponemos los valores iniciales que podemos calcular ahora
	// Valores iniciales del texto de presentacion
	title.x = TITLEX0;
	title.y = TITLEY0;
	title.w = TITLEW;
	title.h = TITLEH;
	title.c = TITLEC;
	// Valor del tiempo de velocidad fija
	d = MAXFIX;

	// Ponemos el tipo de letra para el fondo, imprimimos en consola
	mgsSetFont(scr, 100, title.c, MGS_TRANSPARENT);
	puts("\n Pulsa S para empezar");
	
	// Y empezamos el bucle. Es una sentencia que selecciona el cdigo de cada estado.
	// Al final, comn para todos, se actualiza la pantalla y se sincroniza
	while(1) {
		switch(st) {
			
		// Estado inicial. Muestra un texto que se mueve por la pantalla
		// y espera que se pulse S
		case PREVIOUS:
			// La variable d cuenta el tiempo para cambiar la velocidad.
			// Cuando alcanza el valor mximo la cambiamos y volvemos a poner
			// d a 0. As se tiene un movimiento aleatorio no demasiado brusco
			if (d == MAXFIX) {
				// Las velocidades van entre -2 y 2
				title.vx = -2 + rand() % 5;
				title.vy = -2 + rand() % 5;
				d = 0;
			}
			d++;
			// Incrementamos las coordenadas de acuerdo con la velocidad
			title.x += title.vx;
			title.y += title.vy;
			// Detectamos si nos salimos de la pantalla. En ese caso, corregimos la
			// posicin para seguir dentro y cambiamos la velocidad a 1 en sentido
			// contrario
			if (title.x < 0) {
				title.x = 0;
				title.vx = 1;
			}
			else if (title.x > SCRW - title.w) {
				title.x = SCRW - title.w;
				title.vx = - 1;
			}
			// Lo mismo en y
			if (title.y < 0) {
				title.y = 0;
				title.vy = 1;
			}
			else if (title.y > SCRH - title.h) {
				title.y = SCRH - title.h;
				title.vy = - 1;
			}
			// Limpiamos la pantalla, es decir el texto anterior, y
			// escribimos el texto en la nueva posicin
			mgsClearScreen(scr);
			mgsPuts(scr, title.x, title.y, "Press S to start");
			// Si pulsamos S cambiamos el tamao y color del texto y el tamao para los marcadores,
			// cambiamos de estado e imprimimos unas normas de juego en consola.
			if(mgsKeyPressed('S')) {
				mgsSetFont(scr, 120, TXTBKGC, MGS_TRANSPARENT);
				st = NORMAL;
				// Valores iniciales para el estado NORMAL
				// Valores iniciales para la pelota. La velocidad horizontal es
				// aleatoria entre 3 y -3. La vertical es de subida -negativa- y se
				// calcula para que la suma del mdulo de ambas sea 6. De esta forma se mantiene
				// una velocidad lineal ms o menos uniforme -parecida al mdulo del
				// vector velocidad. La pelota aparece encima y en mitad del bate.
				ball.vx = rand() % 7 - 3;
				ball.vy = -(6 - abs(ball.vx));
				ball.x = BALLX0;
				ball.y = BALLY0;
				ball.h = BALLR;
				ball.w = BALLR;
				ball.c = BALLC;
				// Ponemos los valores iniciales del bate
				bat.vx = INCBAT;
				bat.x = BATX0;
				bat.y = BATY0;
				bat.w = BATW;
				bat.h = BATH;
				bat.c = BATC;
				// Ponemos los valores iniciales de la estrella, con componentes aleatorias
				gift.x = rand() % (SCRW - GIFTER);	// Aparecer en el borde superior, pero en cualquier posicin en horizontal
				gift.y = 0;
				gift.vy = 1 + (rand() & 3); // La velocidad de cada sera entre 1 y 4
				gift.h = GIFTER; // Ancho y alto fijos, iguales al radio exterior
				gift.w = GIFTER;
				gift.extra[GIFTIXM] = rand() & 1;	// Tipo de mejora 0 o 1 aleatorio
				gift.extra[GIFTIXP] = 4 + (rand() & 3); // Entre 4 y 7 puntas
				gift.extra[GIFTIXR] = 0; // Angulo de rotacion inicial
				gift.c = mgsRGBColor(128 + (rand() & 0x7F), 128 + (rand() & 0x7F), 128 + (rand() & 0x7F)); // Color claro aleatorio
				giftTime = GIFTMAX + (rand() & 0x3FF); // Tiempo de aparicin aleatorio entre GIFTMAX y GIFTMAX + 1023

				// Las variables de texto las iniciamos a la cadena "0"
				sRebounds[0] = '0';
				sRebounds[1] = '\0';
				sOut[0] = '0';
				sOut[1] = '\0';
	
				puts("\nUsa <= y => para mover la pala");
				puts("Intenta capturar las estrellas");
				puts("Pulsa ESC cuando te canses");
			}
			// As los numeros aleatorios dependen del tiempo de espera y son aleatorios de verdad
			i = rand();
			break;
		
		// Estado que implementa todo el juego. Mueve la pelota, el bate y calcula la aparicin
		// y movimiento de las estrellas
		case NORMAL:
			// Actualizamos la posicin de la pelota
			ball.x += ball.vx;
			ball.y += ball.vy;
			// En el eje X, en caso de rebote con las paredes ajustamos la posicin
			// y cambiamos el sentido de la velocidad.
			if (ball.x < BHWIDTH + ball.w) {
				ball.x = BHWIDTH + ball.w;
				ball.vx = -ball.vx;
				// Rebote en pared
				mgsPlaySoundFile("wall.wav", 0);
			}
			else if (ball.x > SCRW - BHWIDTH - ball.w) {
				ball.x = SCRW - BHWIDTH - ball.w;
				ball.vx = -ball.vx;
				// Rebote en pared
				mgsPlaySoundFile("wall.wav", 0);
			}
			// En el eje y es ms complicado. Si el rebote es con la pared 
			// superior simplemente ajustamos la Y y cambiamos el sentido
			// de la velocidad. Adems incrementamos el contador de rebotes
			// y actualizamos la cadena para mostrarlos
			if (ball.y < BVWIDTH + ball.h) {
				ball.y = BVWIDTH + ball.h;
				ball.vy = -ball.vy;
				// Rebote en pared
				mgsPlaySoundFile("wall.wav", 0);
				rebound++;
				sprintf(sRebounds, "%d", rebound);
			}
			// En la parte inferior es ms complicado, pues hemos de encontrar o no el bate.
			else if (ball.y > bat.y - ball.h) {
				// Si lo encontramos - y el clculo es con las coordenadas X- ponemos la bola
				// encima del bate y recalculamos la velocidad tras el rebote. 
				// La VX se calcula sumando entre 3 y -3 a la original, segn se rebote ms cerca
				// cerca del centro del bate y hacia la izquierda o la derecha.
				// La nueva VY ser de subida y calculada para mantener ms o menos el mdulo de
				// la velocidad en 6
				if ((ball.x < bat.x + ball.w + bat.w) && (ball.x > bat.x - ball.w)) {
					// Corregimos la posicin Y
					ball.y = bat.y - ball.h;
					// Calculamos distancia al centro del bate
					d = ball.x - bat.x - (bat.w >> 1); 
					// se divide entre la mitad del ancho entre 4 -se multiplica po 8 y se divide entre el ancho
					d = (d << 3) / bat.w;
					// puede ser mayor que 3 en modulo, pues la pelota puede no tocar en el centro
					if (d < -3) d = -3;
					else if (d > 3) d = 3;
					// Ajustamos la velocidad en X
					ball.vx = ball.vx + d;
					if (ball.vx < -3) ball.vx = -3;
					else if (ball.vx > 3) ball.vx = 3;
					// Ajuste de vy
					ball.vy = -(6 - abs(ball.vx));
					// Rebote en el bate
					mgsPlaySoundFile("bat.wav", 0);
				}
				// Si no encontramos el bate, incrementamos el contador de bolas perdidas y preparamos
				// la animacin del estado GOOUT.
				else {
					// Incrementamos el contador y cambiamos el texto asociado
					out++;
					sprintf(sOut, "%d", out);
					// Iniciamos el contador de tiempo de la animacin del estado GOOUT
					animOut = 0;
					// Preparamos los 8 fragmentos en que se "rompe" el bate, con las posiciones
					// adecuadas, tamaos y color fijos, y velocidades dependientes de su posicin
					for (i = 0; i < 8; i++) {
						fragments[i].w = bat.w >> 2;
						fragments[i].h = bat.h >> 1;
						fragments[i].x = bat.x + (i & 3) * fragments[i].w;
						fragments[i].y = bat.y + (i >> 2) * fragments[i].h;
						fragments[i].vx = -4 + ((i & 3) << 1) + (rand() & 3);
						fragments[i].vy = -3 + ((i >> 2) << 1) - (rand() & 3);
					}
					// Comenzamos el sonido repetido de fallo
					mgsPlaySoundFile("out.wav", MGS_LOOP);
					// Y preparamos el cambio de estado
					st = GOOUT;
				}
			}
			
			// Gestionamos la estrella, que est en el mundo cuando su contador de tiempo de
			// espera vale 0
			if (!giftTime) {
				// Actualizamos la posicin de cada segn VY y la horizontal con un valor
				// aleatorio para que se mueva poco, pero errticamente en X
				gift.y += gift.vy;
				gift.x += (rand() % 5) - 2;
				// Rotamos 2 grados antihorario, y ajustamos
				gift.extra[GIFTIXR] += 2;
				if (gift.extra[GIFTIXR] == 360)
					gift.extra[GIFTIXR] = 0;
				// Y vamos al proceso de deteccin de colisiones similar al de la pelota.
				// Primero vemos si ha llegado a la Y del bate
				if (gift.y > bat.y - gift.h) {
					// Y en caso afirmativo vemos si ha colisionado con el bate -ha sido capturada
					if ((gift.x < bat.x + gift.w + bat.w) && (gift.x > bat.x - gift.w)) {
						// De ser as, aplicamos la mejora. Lo hacemos con un switch para que sea
						// fcil aadir ms o modificar -por ejemplo, cambiar el color de bate segn
						// la mejora, etctera
						switch(gift.extra[GIFTIXM]) {
						case GIFTCW:
							bat.w = BATWGIF;
							break;
						case GIFTCV:
							bat.vx = BATVGIF;
							break;
						default:
							break;
						}
						// Cambiamos el color sea cual sea la mejora
						// Ponemos en marcha el contador de tiempo de mejora
						// e incrementamos el numero de estrellas capturadas
						bat.c = BATGIFC;
						widthTime = BATGIFT;
						cStars++;
						// Sonido de estrella capturada
						mgsPlaySoundFile("getStar.wav", 0);
					}
					// Incrementamos el contador de estrellas
					nStars++;
					// Y recalculamos los datos de la nueva, empezando por el retardo hasta la salida
					// Sumamos el tiempo de mejora, que ser el total si ha sido capturada o 0 si no
					giftTime = widthTime + GIFTMAX + (rand() & 0x3FF);
					gift.x = rand() % (SCRW - GIFTER);
					gift.y = 0;
					gift.vy = 1 + (rand() & 3);
					gift.h = GIFTER;
					gift.w = GIFTER;
					gift.extra[GIFTIXM] = rand() & 1;
					gift.extra[GIFTIXP] = 4 + (rand() & 3);
					gift.extra[GIFTIXR] = 0;
					gift.c = mgsRGBColor(128 + (rand() & 0x7F), 128 + (rand() & 0x7F), 128 + (rand() & 0x7F));
				}
			}
			// Si no hay estrella, decrementamos su retardo de aparicin.
			else {
				giftTime--;
				if (!giftTime)
					// Sonido de estrella
					mgsPlaySoundFile("star.wav", 0);
			}
			
			// Tratamos el tiempo de mejora. Si es 0 no hacemos nada.
			if (widthTime) {
				// Si es distinto lo decrementamos y, al llegar a 0
				// volvemos el bate a su estado normal con independencia de la mejora que hubiera
				widthTime--;
				if (!widthTime) {
					bat.w = BATW;
					bat.vx = INCBAT;
					bat.c = BATC;
				}
			}
			
			// Gestionamos las teclas de movimiento y posicin del bate.
			// Segn la tecla, aplicamos la velocidad actual a izquierda o
			// derecha -es decir, restando o sumando- y corregimos para no
			// salir de los borders.
			if (mgsKeyPressed(MGS_KLEFTARROW)) {
				bat.x -= bat.vx;
				if (bat.x < BHWIDTH) bat.x = BHWIDTH;
			}
			if (mgsKeyPressed(MGS_KRIGHTARROW)) {
				bat.x += bat.vx;
				if (bat.x > SCRW - BHWIDTH - bat.w) bat.x = SCRW - BHWIDTH - bat.w;
			}
			// Una vez hechos los clculos, borramos la pantalla y dibujamos los
			// nuevos objetos: los dos textos, la estrella si hay, el bate y la pelota.
			mgsClearScreen(scr);
			mgsPuts(scr, 150, 100, sRebounds);
			mgsPuts(scr, SCRH, 100, sOut);
			if (!giftTime) mgsRotatedStarCR(scr, gift.extra[GIFTIXP], gift.x, gift.y, GIFTER, GIFTIR, gift.extra[GIFTIXR], gift.c);
			mgsRectangle(scr, bat.x, bat.y, bat.x + bat.w, bat.y + bat.h, bat.c);
			mgsEllipse(scr, ball.x, ball.y, ball.w, ball.h, ball.c);
		break;
			
		// Estado que muestra una animacin al perder la pelota.
		// Al terminar vuelve al NORMAL dando valores iniciales al bate y la pelota.
		// Si hubiera una estrella en el mundo, la elimina y calcula valores iniciales
		// para la siguiente.
		// Adems pone a 0 el tiempo de mejora por si hubiera alguna en marcha.
		case GOOUT:
			// Incrementamos el tiempo del estado
			animOut++;
			// Limpiamos la pantalla e imprimimos
			mgsClearScreen(scr);
			// los dos textos
			mgsPuts(scr, 150, 100, sRebounds);
			mgsPuts(scr, SCRH, 100, sOut);
			// Y, para los 8 fragmentos, tambin calculamos su nueva posicin. No verificamos nada, no importa.
			for(i = 0; i < 8; i++) {
				mgsRectangle(scr, fragments[i].x, fragments[i].y, fragments[i].x + fragments[i].w, fragments[i].y + fragments[i].h, MGS_CGREEN);
				fragments[i].x += fragments[i].vx;
				fragments[i].y += fragments[i].vy;
			}
			// Si hemos llegado al final preparamos los datos de la pelota y el bate
			if (animOut == ANIMOUT) {
				ball.vx = rand() % 7 - 3;
				ball.vy = -(6 - abs(ball.vx));
				ball.x = BALLX0;
				ball.y = BALLY0;
				ball.h = BALLR;
				ball.w = BALLR;
				bat.x = BATX0;
				bat.y = BATY0;
				bat.w = BATW;
				bat.h = BATH;
				bat.c = BATC;
				// Si hay alguna estrella recalculamos los valores de la siguiente
				if (!giftTime) {
					// Contamos la estrella, que no lo hemos hecho antes
					nStars++;
					giftTime = GIFTMAX + (rand() & 0x3FF);
					gift.x = rand() % (SCRW - GIFTER);
					gift.y = 0;
					gift.vy = 1 + (rand() & 3);
					gift.h = GIFTER;
					gift.w = GIFTER;
					gift.extra[GIFTIXM] = rand() & 1;
					gift.extra[GIFTIXP] = 4 + (rand() & 3);
					gift.extra[GIFTIXR] = 0;
					gift.c = mgsRGBColor(128 + (rand() & 0x7F), 128 + (rand() & 0x7F), 128 + (rand() & 0x7F));
				}
				// Eliminamos las mejoras
				widthTime = 0;
				// Paramos el sonido de fallo
				mgsStopSound();
				// Y volvemos al estado normal
				st = NORMAL;
			}
		break;
		}
		
		// Para todos los estados, verificamos si se ha pulsado ESC
		if (mgsKeyPressed(MGS_KESC)) {
			// De ser as, borramos la pantalla, imprimimos las estdisticas
			mgsDeleteScreen(scr);
			printf("\nHas conseguido %d rebotes, has dejado pasar %d bolas.\n", rebound, out);
			printf("Han salido %d estrellas y has capturado %d.\nHasta otra.\n\n", nStars, cStars);
			puts("\n Pulsa T para terminar");
			// Y esperamos a pulsar T para terminar
			while(!mgsKeyPressed('T'));
			exit(EXIT_SUCCESS);
		}
		// Si no hemos de salir, actualizamos la pantalla con los nuevos objetos
		mgsUpdateScreen(scr);
		// Y esperamos a que se cumpla el tiempo de frame para la animacin
		mgsWaitFrame();
	}
}

